#!/usr/bin/env python3

"""Supporting utilities.

Classes
-------
.. autosummary::
    ProgressBar
    OptionReader

Routines
--------
.. autosummary::
    read_param
    round_up
    evaluate_ratio
    humansize
    humantime

----

"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

try:
    import configparser
except ImportError:
    import ConfigParser as configparser
import math
import os
import re
import sys
import time


def read_param(params, key, default):
    """Read and return a parameter from a dict.

    If the key `key` is absent from the dict `params`, return the
    default value `default` instead.

    Parameters
    ----------
    params : dict
        A dict containing parameters.
    key : str
        Name of the parameter, i.e., its corresponding key in `params`.
    default
        Default value for the parameter, if `key` is absent from `params`.

    Returns
    -------
    value
        If `key` in `params`, return ``params[key]``; otherwise, return
        `default` instead.

    """

    if not isinstance(key, str):
        raise ValueError('invalid parameter name %s' % str(key))

    return params[key] if key in params else default


def round_up(number, ndigits=0):
    """Round a floating point number *upward* to a given precision.

    Unlike the builtin `round`, the return value `round_up` is always
    the smallest float *greater than or equal to* the given number
    matching the specified precision.

    Parameters
    ----------
    number : float
        Number to be rounded up.
    ndigits : int, optional
        Number of decimal digits in the result. Default is 0.

    Returns
    -------
    float

    Examples
    --------
    >>> round_up(math.pi)
    4.0
    >>> round_up(math.pi, ndigits=1)
    3.2
    >>> round_up(math.pi, ndigits=2)
    3.15
    >>> round_up(-math.pi, ndigits=4)
    -3.1415

    """

    multiplier = 10 ** ndigits
    return math.ceil(number * multiplier) / multiplier


# patterns numerator:denominator and numerator/denominator
_NUM_COLON_DEN = re.compile(r'^([1-9][0-9]*):([1-9][0-9]*)$')
_NUM_SLASH_DEN = re.compile(r'^([1-9][0-9]*)/([1-9][0-9]*)$')


def evaluate_ratio(ratio_str):
    """Evaluate ratio in the form num:den or num/den.

    Note that numerator and denominator should both be positive
    integers.

    Parameters
    ----------
    ratio_str : str
        The ratio as a string (either ``'num:den'`` or ``'num/den'``
        where ``num`` and ``den``, the numerator and denominator, are
        positive integers.

    Returns
    -------
    ratio : float
        The ratio as a float, or ``None`` if `ratio_str` is malformed.

    Examples
    --------
    >>> evaluate_ratio('16:9')
    1.7777777777777777
    >>> evaluate_ratio('16/9')
    1.7777777777777777
    >>> print(evaluate_ratio('0/9'))
    None

    """

    match = _NUM_COLON_DEN.match(ratio_str)
    if match:
        numerator = int(match.group(1))
        denominator = int(match.group(2))
        return numerator / denominator
    match = _NUM_SLASH_DEN.match(ratio_str)
    if match:
        numerator = int(match.group(1))
        denominator = int(match.group(2))
        return numerator / denominator
    return None


def humansize(size):
    """Return a human readable string of the given size in bytes."""
    multiplier = 1024.0
    if size < multiplier:
        return "%dB" % size
    for unit in ['Ki', 'Mi', 'Gi', 'Ti', 'Pi', 'Ei', 'Zi']:
        size /= multiplier
        if size < multiplier:
            if size < 10:
                return "%.2f%sB" % (round_up(size, 2), unit)
            elif size < 100:
                return "%.1f%sB" % (round_up(size, 1), unit)
            else:
                return "%.0f%sB" % (round_up(size, 0), unit)
            break
    else:
        return "%.1f%sB" % (round_up(size, 1), unit)


def humantime(seconds, ndigits=2, one_hour_digit=False):
    """Format a duration as a human readable string.

    The duration in seconds (a nonnegative float) is formatted as
    ``HH:MM:SS.frac``, where the number of fractional digits is
    controlled by `ndigits`; if `ndigits` is 0, the decimal point is not
    printed. The number of hour digits (``HH``) can be reduced to one
    with the `one_hour_digits` option.

    Parameters
    ----------
    seconds : float
        Duration in seconds, must be nonnegative.
    ndigits : int, optional
        Number of digits after the decimal point for the seconds part.
        Default is 2. If 0, the decimal point is suppressed.
    one_hour_digit : bool, optional
        If ``True``, only print one hour digit (e.g., nine hours is
        printed as 9:00:00.00). Default is ``False``, i.e., two hour
        digits (nine hours is printed as 09:00:00.00).

    Returns
    -------
    human_readable_duration : str

    Raises
    ------
    ValueError:
        If `seconds` is negative.

    Examples
    --------
    >>> humantime(10.55)
    '00:00:10.55'
    >>> humantime(10.55, ndigits=1)
    '00:00:10.6'
    >>> humantime(10.55, ndigits=0)
    '00:00:11'
    >>> humantime(10.55, one_hour_digit=True)
    '0:00:10.55'
    >>> # two hours digits for >= 10 hours, even if one_hour_digit is
    >>> # set to True
    >>> humantime(86400, one_hour_digit=True)
    '24:00:00.00'
    >>> humantime(-1)
    Traceback (most recent call last):
        ...
    ValueError: seconds=-1.000000 is negative, expected nonnegative value

    """

    # pylint: disable=invalid-name
    if seconds < 0:
        raise ValueError("seconds=%f is negative, "
                         "expected nonnegative value" % seconds)

    hh = int(seconds) // 3600  # hours
    mm = (int(seconds) // 60) % 60  # minutes
    ss = seconds - (int(seconds) // 60) * 60  # seconds
    hh_str = "%01d" % hh if one_hour_digit else "%02d" % hh
    mm_str = "%02d" % mm
    if ndigits == 0:
        ss_str = "%02d" % round(ss)
    else:
        ss_format = "%0{0}.{1}f".format(ndigits + 3, ndigits)
        ss_str = ss_format % ss
    return "%s:%s:%s" % (hh_str, mm_str, ss_str)


# default progress bar update interval
_PROGRESS_UPDATE_INTERVAL = 1.0
# the format string for a progress bar line
#
# 0: processed size, e.g., 2.02GiB
# 1: elapsed time (7 chars), e.g., 0:00:04
# 2: current processing speed, e.g., 424MiB (/s is already hardcoded)
# 3: the bar, in the form "=====>   "
# 4: number of percent done, e.g., 99
# 5: estimated time remaining (11 chars), in the form "ETA H:MM:SS"; if
#    finished, fill with space
_FORMAT_STRING = '\r{0:>7s} {1} [{2:>7s}/s] [{3}] {4:>3s}% {5}'


class ProgressBar(object):
    """Progress bar for file processing.

    To generate a progress bar, init a ProgressBar instance, then update
    frequently with the `update` method, passing in the size of newly
    processed chunk. The `force_update` method should only be called if
    you want to overwrite the processed size, which is automatically
    calculated incrementally. After you finish processing the
    file/stream, you must call the `finish` method to wrap it up. Any
    further calls after the `finish` method has been called lead to
    a ``RuntimeError``.

    Each ProgressBar instance defines several public attributes listed
    below. Some are available during processing, and some after
    processing. These attributes are meant for informational purposes,
    and you should not manually tamper with them (which mostly likely
    leads to undefined behavior).

    The progress bar format is inspired by ``pv(1)`` (pipe viewer).

    Parameters
    ----------
    totalsize : int
        Total size, in bytes, of the file/stream to be processed.
    interval : float, optional
        Update (refresh) interval of the progress bar, in
        seconds. Default is 1.0.

    Attributes
    ----------
    totalsize : int
        Total size of file/stream, in bytes. Available throughout.
    processed : int
        Process size. Available only during processing (deleted after
        the `finish` call).
    start : float
        Starting time (an absolute time returned by
        ``time.time()``). Available throughout.
    interval : float
        Update (refresh) interval of the progress bar, in
        seconds. Available only during processing (deleted after the
        `finish` call).
    elapsed : float
        Total elapsed time, in seconds. Only available after the
        `finish` call.

    Notes
    -----
    For developers: ProgressBar also defines three private attributes,
    `_last`, `_last_processed` and `_barlen`, during processing (deleted
    after the `finish` call). `_last` stores the absolute time of last
    update (refresh), `_last_processed` stores the processed size at the
    time of the last update (refresh), and `_barlen` stores the length
    of the progress bar (only the bar portion).

    There is another private attribute `__finished` (bool) keeping track
    of whether `finish` has been called. (Protected with double leading
    underscores since no one should ever tamper with this.)

    """

    # pylint: disable=too-many-instance-attributes

    def __init__(self, totalsize, interval=_PROGRESS_UPDATE_INTERVAL):
        """Initialize the ProgressBar class.

        See class docstring for parameters of the constructor.

        """

        self.totalsize = totalsize
        self.processed = 0
        self.start = time.time()
        self.interval = interval
        self._last = self.start
        self._last_processed = 0
        self.__finished = False

        # calculate bar length
        try:
            ncol, _ = os.get_terminal_size()
        except (AttributeError, OSError):
            # Python2 do not have os.get_terminal_size.  Also,
            # os.get_terminal_size fails if stdout is redirected to a
            # pipe (pretty stupid -- should check stderr; relevant
            # Python bug: https://bugs.python.org/issue14841). In either
            # case, Assume a minimum of 80 columns.
            ncol = 80
        self._barlen = (ncol - 48) if ncol >= 58 else 10

    def update(self, chunk_size):
        """Update the progress bar for a newly processed chunk.

        The size of the processed chunk is registered. Whether the
        progress bar is refreshed depends on whether we have reached the
        refresh interval since the last refresh (handled automatically).

        Parameters
        ----------
        chunk_size : int
            The size of the newly processed chunk (since last update),
            in bytes. This size will be added to the `processed`
            attribute.

        Raises
        ------
        RuntimeError:
            If `finish` has been called on the ProgressBar instance.

        """

        if self.__finished:
            raise RuntimeError('operation on finished progress bar')

        self.processed += chunk_size
        if self.processed > self.totalsize:
            self.processed = self.totalsize
        self._update_output()

    def force_update(self, processed_size):
        """Force update the progress bar with a given processed size.

        The `processed` attribute is overwritten by the new value.

        Parameters
        ----------
        processed_size :
            Processed size of the file/stream, in bytes. Existing value
            is overwritten by this value.

        Raises
        ------
        RuntimeError:
            If `finish` has been called on the ProgressBar instance.

        """

        if self.__finished:
            raise RuntimeError('operation on finished progress bar')

        self.processed = processed_size
        if self.processed > self.totalsize:
            self.processed = self.totalsize
        self._update_output()

    def finish(self):
        """Finish file progressing and wrap up on the progress bar.

        Always call this method exactly once after you finish
        processing. This method adds the finishing touches to the
        progress bar, deletes several attributes (`processed`,
        `interval`), and adds a new attribute (`elapsed`).

        After `finish` is called on a ProgressBar attribute, it enters a
        read-only mode: you may read the `totalsize`, `start`, and
        `elapsed` attributes, but any method call leads to a
        ``RuntimeError``.

        Raises
        ------
        RuntimeError:
            If `finish` has already been called on the ProgressBar
            instance before.

        """

        # pylint: disable=attribute-defined-outside-init

        if self.__finished:
            raise RuntimeError('operation on finished progress bar')

        self.elapsed = time.time() - self.start
        if self.elapsed < 0.001:
            self.elapsed = 0.001  # avoid division by zero
        del self.processed
        del self.interval
        del self._last
        del self._last_processed

        self.__finished = True

        processed_s = humansize(self.totalsize)
        elapsed_s = self._humantime(self.elapsed)
        speed_s = humansize(self.totalsize / self.elapsed)
        bar_s = '=' * (self._barlen - 1) + '>'
        percent_s = '100'
        eta_s = ' ' * 11
        sys.stderr.write(_FORMAT_STRING.format(
            processed_s, elapsed_s, speed_s, bar_s, percent_s, eta_s
        ))
        sys.stderr.write("\n")
        sys.stderr.flush()

    def _update_output(self):
        """Update the progress bar and surrounding data as appropriate.

        Whether the progress bar is refreshed depends on whether we have
        reached the refresh interval since the last refresh (handled
        automatically).

        Raises
        ------
        RuntimeError:
            If `finish` has already been called on the ProgressBar
            instance before.

        """

        if self.__finished:
            raise RuntimeError('operation on finished progress bar')

        elapsed_since_last = time.time() - self._last
        if elapsed_since_last < self.interval:
            return

        if elapsed_since_last < 0.001:
            elapsed_since_last = 0.001  # avoid division by zero

        # speed in the last second, in bytes per second
        speed = ((self.processed - self._last_processed) / elapsed_since_last)

        # update last stats for the next update
        self._last = time.time()
        self._last_processed = self.processed

        # _s suffix stands for string
        processed_s = humansize(self.processed)
        elapsed_s = self._humantime(time.time() - self.start)
        speed_s = humansize(speed)
        percentage = self.processed / self.totalsize  # absolute
        percent_s = str(int(percentage * 100))
        # generate bar
        length = int(round(self._barlen * percentage))
        fill = self._barlen - length
        if length == 0:
            bar_s = " " * self._barlen
        else:
            bar_s = '=' * (length - 1) + '>' + ' ' * fill
        # calculate ETA
        remaining = self.totalsize - self.processed
        # estimate based on current speed
        eta = remaining / speed
        eta_s = "ETA %s" % self._humantime(eta)

        sys.stderr.write(_FORMAT_STRING.format(
            processed_s, elapsed_s, speed_s, bar_s, percent_s, eta_s
        ))
        sys.stderr.flush()

    @staticmethod
    def _humantime(seconds):
        """Customized humantime for ProgressBar."""
        return humantime(seconds, ndigits=0, one_hour_digit=True)


class OptionReader(object):
    """Class for reading options from a list of fallbacks.

    OptionReader optionally takes command line arguments parsed by
    argparse, a list of possible configuration files, and a dictionary
    of default values. Then one can query the class for the value of an
    option using the ``opt(name)`` method. The value is determined in
    the order of CLI argument, value specified in config files, default
    value, and at last ``None`` if none of the above is available.

    Parameters
    ----------
    cli_args : argparse.Namespace, optional
        CLI arguments returned by
        ``argparse.ArgumentParser.parse_args()``. If ``None``, do not
        consider CLI arguments. Default is ``None``.
    config_files : str or list, optional
        Path(s) to the expected configuration file(s). If ``None``,
        do not read from config files. Default is ``None``.
    section : str, optional
        Name of the config file section to read from. Do not use
        ``DEFAULT``, as it is the reserved name for a special
        section. If ``None``, do not read from config files. Default is
        ``None``.
    defaults : dict, optional
        A dict containing default values of one or more options. If
        ``None``, do not consider default values. Default is ``None``.

    Raises
    ------
    configparser.Error:
        If some of the supplied configuration files are malformed.

    Notes
    -----
    For developers: there are three private attributes, ``_cli_opts``,
    ``_cfg_opts`` and ``_default_opts``, which are dicts containing CLI,
    config file, and default options, respectively.

    """

    def __init__(self, cli_args=None, config_files=None, section=None,
                 defaults=None):
        """
        Initialize the OptionReader class.

        See class docstring for parameters.

        """

        if section == 'DEFAULT':
            raise ValueError("section name DEFAULT is not allowed")

        # parse CLI arguments
        if cli_args is not None:
            # only include values that are not None
            self._cli_opts = dict((k, v) for k, v in cli_args.__dict__.items()
                                  if v is not None)
        else:
            self._cli_opts = {}

        # parse config files
        if config_files is not None and section is not None:
            config = configparser.ConfigParser()
            config.read(config_files)
            if config.has_section(section):
                self._cfg_opts = dict(config.items(section))
            else:
                self._cfg_opts = {}
        else:
            self._cfg_opts = {}

        # default options
        if defaults is not None:
            self._default_opts = defaults
        else:
            self._default_opts = {}

    def cli_opt(self, name):
        """
        Read the value of an option from the corresponding CLI argument.

        Parameters
        ----------
        name : str
            Name of the option.

        Returns
        -------
        value
            Value of the corresponding CLI argument if available, and
            ``None`` otherwise.

        """

        return self._cli_opts[name] if name in self._cli_opts else None

    def cfg_opt(self, name, opttype=None):
        """
        Read the value of an option from config files.

        Parameters
        ----------
        name : str
            Name of the option.
        opttype : {None, str, int, float, bool}
            Type of the option. The value of the option is converted to
            the corresponding type. If ``None``, no conversion is done
            (so the return type will actually be ``str``). If ``bool``,
            the returned value will be ``True`` if the default value is
            ``'yes'``, ``'on'``, or ``'1'``, and ``False`` if the
            default value is ``'no'``, ``'off'``, or ``'0'`` (case
            insensitive), just like
            ``configparser.ConfigParser.getboolean``.

        Returns
        -------
        value
            Value of the option in config files if available, and
            ``None`` otherwise.

        Raises
        ------
        ValueError:
            If the raw value of the option in config files (if
            available) cannot be converted to `opttype`, or if `opttype`
            is not one of the supported types.

        """

        # pylint: disable=too-many-return-statements

        if name not in self._cfg_opts:
            return None

        rawopt = self._cfg_opts[name]
        if opttype is None:
            return rawopt
        elif opttype is str:
            return rawopt
        elif opttype is int:
            return int(rawopt)
        elif opttype is float:
            return float(rawopt)
        elif opttype is bool:
            rawopt_lower = rawopt.lower()
            if rawopt_lower in {'yes', 'on', '1'}:
                return True
            elif rawopt_lower in {'no', 'off', '0'}:
                return False
            else:
                raise ValueError("not a boolean: %s" % rawopt)
        else:
            raise ValueError("unrecognized opttype %s" % str(opttype))

    def default_opt(self, name):
        """
        Read the default value of an option.

        Parameters
        ----------
        name : str
            Name of the option.

        Returns
        -------
        value
            Default value of the option if available, and ``None`` otherwise.

        """

        return self._default_opts[name] if name in self._default_opts else None

    def opt(self, name, opttype=None):
        """
        Read the value of an option.

        The value is determined in the following order: CLI argument,
        config files, default value, and at last ``None``.

        Parameters
        ----------
        name : str
            Name of the option.
        opttype : {None, str, int, float, bool}
            Type of the option, only useful for config files. See the
            `opttype` parameter of the `cfg_opt` method.

        Returns
        -------
        value
            Value of the option.

        Raises
        ------
        ValueError:
            If the raw value of the option in config files (if
            available) cannot be converted to `opttype`, or if `opttype`
            is not one of the supported types.

        """

        if name in self._cli_opts:
            return self._cli_opts[name]
        elif name in self._cfg_opts:
            return self.cfg_opt(name, opttype)
        elif name in self._default_opts:
            return self._default_opts[name]
        else:
            return None



#!/usr/bin/env python3

"""Extract video metadata with FFprobe.

Classes
-------
.. autosummary::
    Stream
    Video

Routines
--------
.. autosummary::
    main

----

"""

# pylint: disable=too-many-lines

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import argparse
import fractions
import hashlib
import json
import os
import subprocess
import sys

from storyboard import fflocate
from storyboard import util
from storyboard.util import read_param as _read_param
from storyboard import version


_FORMAT_MAP = {
    'aac': 'Raw ADTS AAC',
    'ac3': 'Raw AC-3',
    'aiff': 'Audio Interchange File Format (AIFF)',
    'asf': 'Advanced Systems Format',
    'avi': 'Audio Video Interleaved',
    'flac': 'Native FLAC',
    'flv': 'Flash video',
    'jpeg_pipe': 'JPEG',
    'matroska,webm': 'Matroska',
    'mp3': 'MP3',
    'mpeg': 'MPEG program stream',
    'mpegts': 'MPEG transport stream',
    'mpegvideo': 'Raw MPEG video',
    'mov,mp4,m4a,3gp,3g2,mj2': 'MPEG-4 Part 14',
    'ogg': 'Ogg',
    'rm': 'RealMedia',
    'png_pipe': 'PNG',
}

_VCODEC_MAP = {
    'h264': 'H.264',
    'hevc': 'HEVC',
    'mjpeg': 'Motion JPEG',
    'mpeg1video': 'MPEG-1 Part 2',
    'mpeg2video': 'MPEG-2 Part 2',
    'mpeg4': 'MPEG-4 Part 2',
    'png': 'PNG',
    'rv10': 'RealVideo 1.0',
    'rv20': 'RealVideo 2.0',
    'rv30': 'RealVideo 3.0',
    'rv40': 'RealVideo 4.0',
    'theora': 'Theora',
    'vp8': 'VP8',
    'vp9': 'VP9',
}

_ACODEC_MAP = {
    'aac': 'AAC',
    'ac3': 'Dolby AC-3',
    'cook': 'Cook (RealAudio G2)',
    'flac': 'FLAC (Free Lossless Audio Codec)',
    'mp3': 'MP3',
    'real_144': 'RealAudio 1.0',
    'real_288': 'RealAudio 2.0',
    'ralf': 'RealAudio Lossless',
    'vorbis': 'Vorbis',
}

_SCODEC_MAP = {
    'ass': 'SubStation Alpha',
    'cc_dec': 'closed caption (EIA-608 / CEA-708)',
    'srt': 'SubRip',
}


class Stream(object):

    """Container for stream metadata.

    Some of the documented attributes are optional; in particular, some
    attributes are video stream specific. (In that case of an
    unavailable attribute, the value will be ``None``)

    Attributes
    ----------
    index : int
        Stream index within the video container.

    type : str
        ``'video'``, ``'audio'``, ``'subtitle'``, ``'data'``, etc.

    codec : str
        (Long) name of codec.

    bit_rate : float
        Bit rate of stream, in kb/s.

    bit_rate_text : str
        Bit rate as a human readable string, e.g., ``'360 kb/s'``.

    language_code : str

    width : int

    height : int

    dimension : tuple
        ``(width, height)``, e.g., ``(1920, 1080)``.

    dimension_text : str
        Dimension as a human readable string, e.g. ``'1920x1080'``.

    frame_rate : float
        Frame rate of video stream, in frames per second (fps).

    frame_rate_text : str
        Frame rate as a human readable string, e.g., ``'24 fps'``.

    dar : float
        Display aspect ratio.

    dar_text : str
        Display aspect ratio as a human readable string, e.g.,
        ``'16:9'``.

    info_string : str
        Assembled string of stream metadata, intended for printing.

    """

    # pylint: disable=too-many-instance-attributes,too-few-public-methods
    # a stream can have any number of attributes

    def __init__(self):
        """Initialize the Stream class."""
        # general stream attributes
        self.index = None
        self.type = None
        self.codec = None
        self.bit_rate = None
        self.bit_rate_text = None
        self.language_code = None
        # video stream specific attributes
        self.width = None
        self.height = None
        self.dimension = None
        self.dimension_text = None
        self.frame_rate = None
        self.frame_rate_text = None
        self.dar = None  # display aspect ratio
        self.dar_text = None
        # assembled
        self.info_string = None


class Video(object):

    """Container for video and streams metadata.

    A ``Video`` object holds video metadata (including container
    metadata and per-stream metadata), and generates a formatted string
    for printing upon request (the `format_metadata` method).

    Some of the documented attributes might not be available for certain
    video files (in that case the value will be ``None``).

    Parameters
    ----------
    video : str
        Path to the video file.
    params : dict, optional
        Optional parameters enclosed in a dict. Default is ``None``. See
        the "Other Parameters" section for understood key/value pairs.

    Raises
    ------
    OSError
        If fails to extract metadata with ffprobe, e.g., if the file is
        not present, or in a format that is not recognized by ffprobe,
        or if ffprobe cannot be called, etc.

    Other Parameters
    ----------------
    ffprobe_bin : str, optional
        Name/path of the ffprobe binary (should be callable). By default
        the name is guessed based on OS type. (See the
        storyboard.fflocate module.)
    video_duration : float, optional
        Duration of the video in seconds. If ``None``, extract the
        duration from container metadata. Default is ``None``. This is
        only needed in edge cases where the duration of the video cannot
        be read off from container metadata, or the duration extracted
        is wrong. See `#3
        <https://github.com/zmwangx/storyboard/issues/3>`_ for details.
    print_progress : bool, optional
        Whether to print progress information (to stderr). Default is
        False.
    debug : bool, optional
        Print extra debug information. Default is False.

    Attributes
    ----------
    format : str
        (Long) name of container format.

    title : str

    size : int
        Size of video file in bytes.

    size_text : str
        Size as a human readable string, e.g., ``'128MiB'``.

    duration : float
        Duration of video in seconds.

    duration_text : str
        Duration as a human readable string, e.g., ``'00:02:53.33'``.

    scan_type : str
        ``'Progressive scan'``, ``'Interlaced scan'``, or ``'Telecined
        video'``.

    dimension : (width, height)
        E.g., ``(1920, 1080)``.

    dimension_text : str
        Dimension as a human readable string, e.g. ``'1920x1080'``.

    sha1sum : str
        The SHA-1 hex digest of the video file (40 character hexadecimal
        string). Since computing SHA-1 digest is an expensive operation,
        this attribute is only calculated and set upon request, either
        through `compute_sha1sum` or `format_metadata` with the
        ``include_sha1sum`` optional parameter set to ``True``.

    frame_rate : float
        Frame rate of video stream, in frames per second (fps).

    frame_rate_text : str
        Frame rate as a human readable string, e.g., ``'24 fps'``.

    dar : float
        Display aspect ratio.

    dar_text : str
        Display aspect ratio as a human readable string, e.g.,
        ``'16:9'``.

    streams : list
        A list of Stream objects, containing per-stream metadata.

    Notes
    -----
    The unmodified JSON output of ``ffprobe -show_format -show_streams``
    on the video is saved in a private instance attribute `_ffprobe`.

    """

    # pylint: disable=too-many-instance-attributes
    # again, a video can have any number of metadata attributes

    def __init__(self, video, params=None):
        """Initialize the Video class.

        See class docstring for parameters of the constructor.

        """

        if params is None:
            params = {}
        if 'debug' in params and params['debug']:
            self.__debug = True
        self.__dp("entered StoryBoard.__init__")
        if 'ffprobe_bin' in params:
            ffprobe_bin = params['ffprobe_bin']
        else:
            _, ffprobe_bin = fflocate.guess_bins()
        video_duration = _read_param(params, 'video_duration', None)
        print_progress = _read_param(params, 'print_progress', False)

        self.path = os.path.abspath(video)
        if not os.path.exists(self.path):
            raise OSError("'" + video + "' does not exist")
        self.filename = os.path.basename(self.path)
        if hasattr(self.filename, 'decode'):
            # python2 str, need to be decoded to unicode for proper
            # printing
            self.filename = self.filename.decode('utf-8')

        if print_progress:
            sys.stderr.write("Processing %s\n" % self.filename)
            sys.stderr.write("Crunching metadata...\n")

        self._call_ffprobe(ffprobe_bin)

        self.title = self._get_title()
        self.format = self._get_format()
        self.size, self.size_text = self._get_size()
        if video_duration is None:
            self.duration, self.duration_text = self._get_duration()
        else:
            self.duration = video_duration
            self.duration_text = util.humantime(video_duration)
        self.sha1sum = None  # SHA-1 digest is generated upon request

        # the remaining attributes will be dynamically set when parsing
        # streams
        self.dimension = None
        self.dimension_text = None
        self.frame_rate = None
        self.frame_rate_text = None
        self.dar = None
        self.dar_text = None

        self._process_streams()

        # detect if the file contains any video streams at all and try
        # to extract scan type only if it does
        for stream in self.streams:
            if stream.type == 'video':
                break
        else:
            # no video stream
            self.scan_type = None
            self.__dp("left StoryBoard.__init__")
            return
        self.scan_type = self._get_scan_type(ffprobe_bin, print_progress)
        self.__dp("left StoryBoard.__init__")

    def format_metadata(self, params=None):
        """Return video metadata in one formatted string.

        Parameters
        ----------
        params : dict, optional
            Optional parameters enclosed in a dict. Default is ``None``.
            See the "Other Parameters" section for understood key/value
            pairs.

        Returns
        -------
        str
            A formatted string loaded with video and per-stream
            metadata, which can be printed directly. See the "Examples"
            section for a printed example.

        Other Parameters
        ----------------
        include_sha1sum : bool, optional
            Whether to include the SHA-1 hex digest. Default is
            False. Keep in mind that computing SHA-1 digest is an
            expensive operation, and hence is only performed upon
            request.
        print_progress : bool, optional
            Whether to print progress information (to stderr). Default
            is False.

        Examples
        --------
        >>> import os
        >>> import tempfile
        >>> import requests
        >>> video_uri = 'https://dl.bintray.com/zmwangx/pypi/sample-h264-aac-srt.mkv'
        >>> tempdir = tempfile.mkdtemp()
        >>> video_file = os.path.join(tempdir, 'sample-h264-aac-srt.mkv')
        >>> r = requests.get(video_uri, stream=True)
        >>> with open(video_file, 'wb') as fd:
        ...     for chunk in r.iter_content(65536):
        ...         bytes_written = fd.write(chunk)
        >>> print(Video(video_file).format_metadata({'include_sha1sum': True}))
        Title:                  Example video: H.264 + AAC + SRT in Matroska container
        Filename:               sample-h264-aac-srt.mkv
        File size:              4842 (4.73KiB)
        SHA-1 digest:           95e7d9f9359d8d7ba4ec441bc8cb3830a58ee102
        Container format:       Matroska
        Duration:               00:00:02.08
        Pixel dimensions:       128x72
        Display aspect ratio:   16:9
        Scan type:              Progressive scan
        Frame rate:             25 fps
        Streams:
            #0: Video, H.264 (High Profile level 1.0), 128x72 (DAR 16:9), 25 fps
            #1: Audio (und), AAC (Low-Complexity)
            #2: Subtitle, SubRip
        >>> os.remove(video_file)
        >>> os.rmdir(tempdir)
        """

        self.__dp("entered StoryBoard.format_metadata")
        if params is None:
            params = {}
        include_sha1sum = _read_param(params, 'include_sha1sum', False)
        print_progress = _read_param(params, 'print_progress', False)

        lines = []  # holds the lines that will be joined in the end
        # title
        if self.title:
            lines.append("Title:                  %s" % self.title)
        # filename
        lines.append("Filename:               %s" % self.filename)
        # size
        lines.append("File size:              %d (%s)" %
                     (self.size, self.size_text))
        # sha1sum
        if include_sha1sum:
            self._get_sha1sum(print_progress)
            lines.append("SHA-1 digest:           %s" % self.sha1sum)
        # container format
        lines.append("Container format:       %s" % self.format)
        # duration
        if self.duration_text:
            lines.append("Duration:               %s" % self.duration_text)
        else:
            lines.append("Duration:               Not available")
        # dimension
        if self.dimension_text:
            lines.append("Pixel dimensions:       %s" % self.dimension_text)
        # aspect ratio
        if self.dar_text:
            lines.append("Display aspect ratio:   %s" % self.dar_text)
        # scanning type
        if self.scan_type:
            lines.append("Scan type:              %s" % self.scan_type)
        # frame rate
        if self.frame_rate:
            lines.append("Frame rate:             %s" % self.frame_rate_text)
        # streams
        lines.append("Streams:")
        for stream in self.streams:
            lines.append("    #%d: %s" % (stream.index, stream.info_string))
        self.__dp("left StoryBoard.format_metadata")
        return '\n'.join(lines).strip()

    def compute_sha1sum(self, params=None):
        """Computes the SHA-1 digest of the video file.

        Parameters
        ----------
        params : dict, optional
            Optional parameters enclosed in a dict. Default is ``None``.
            See the "Other Parameters" section for understood key/value
            pairs.

        Returns
        -------
        sha1sum : str
            The SHA-1 hex digest of the video file (40 character
            hexadecimal string).

        Other Parameters
        ----------------
        print_progress : bool, optional
            Whether to print progress information (to stderr). Default
            is False.

        Notes
        -----
        Since computing SHA-1 digest is an expensive operation, the
        digest is only calculated and set upon request, either through
        this method or `format_metadata` with the ``include_sha1sum``
        optional parameter set to ``True``. Further requests load the
        calculated value rather than repeat the computation.

        """

        self.__dp("entered StoryBoard.compute_sha1sum")
        if params is None:
            params = {}
        print_progress = _read_param(params, 'print_progress', False)

        self.__dp("left StoryBoard.compute_sha1sum")
        return self._get_sha1sum(print_progress=print_progress)

    def _call_ffprobe(self, ffprobe_bin):
        """Call ffprobe to extract video metadata.

        ffprobe is called with the -show_format and -show_streams
        options, and its JSON output is parsed and stored in the
        `_ffprobe` attribute.

        Parameters
        ----------
        ffprobe_bin : str
            Name/path of the ffprobe binary (should be callable).

        Raises
        ------
        OSError
            If the ffprobe call returns with nonzero status.

        """

        self.__dp("entered StoryBoard._call_ffprobe")
        ffprobe_args = [
            ffprobe_bin,
            '-print_format', 'json',
            '-show_format', '-show_streams',
            '-hide_banner',
            self.path
        ]
        proc = subprocess.Popen(ffprobe_args,
                                stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        ffprobe_out, ffprobe_err = proc.communicate()
        ffprobe_out = ffprobe_out.decode('utf-8', 'ignore')
        ffprobe_err = ffprobe_err.decode('utf-8', 'ignore')

        self.__dp("ffprobe stdout:")
        self.__dp(ffprobe_out)
        self.__dp("ffprobe stderr:")
        self.__dp(ffprobe_err)
        if proc.returncode != 0:
            msg = ("ffprobe failed on '%s'\nffprobe error message:\n%s"
                   % (self.path, ffprobe_err.strip()))
            raise OSError(msg)
        self._ffprobe = json.loads(ffprobe_out)
        self.__dp("left StoryBoard._call_ffprobe")

    def _get_title(self):
        """Get title of video (if any).

        Returns
        -------
        title : str
            Or ``None`` if not present.

        Notes
        -----
        The title, if present, is stored in ``.format.tags.title`` of
        FFprobe's JSON output.

        """

        self.__dp("entered StoryBoard._get_title")
        video_container_metadata = self._ffprobe['format']
        if (('tags' in video_container_metadata and
             'title' in video_container_metadata['tags'])):
            title = video_container_metadata['tags']['title']
        else:
            self.__dp("left StoryBoard._get_title")
            return None
        self.__dp("left StoryBoard._get_title")
        return title

    def _get_format(self):
        """Get container format of the video.

        Returns
        -------
        format : str
            e.g., ``"MPEG-4 Part 14 (MP4)"``, ``"MPEG transport
            stream"``, ``"Matroska"``

        Notes
        -----
        The container format is stored in ``.format.format_name`` and
        ``.format.format_long_name`` in FFprobe's JSON output. Both the
        short names and long names returned by FFprobe are usually not
        very satisfactory, so we roll our own names for common formats.

        """

        # pylint: disable=too-many-branches

        self.__dp("entered StoryBoard._get_format")
        format_name = self._ffprobe['format']['format_name']
        # lowercase extension without period
        extension = os.path.splitext(self.path)[1].lower()[1:]

        if format_name in _FORMAT_MAP:
            fmt = _FORMAT_MAP[format_name]

            # some formats that require special treatment (i.e., a
            # collection of formats sharing a common format_name, in
            # which case further subdivision is required)
            if format_name == 'mov,mp4,m4a,3gp,3g2,mj2':
                if extension in ['mov', 'qt']:
                    fmt = "QuickTime movie"
                elif extension in ['3gp']:
                    fmt = "3GPP"
                elif extension in ['3g2']:
                    fmt = "3GPP2"
                elif extension in ['mj2', 'mjp2']:
                    fmt = "Motion JPEG 2000"
                else:
                    # mp4, m4v, m4a, etc.
                    fmt = "MPEG-4 Part 14 (%s)" % extension.upper()
            elif format_name == 'matroska,webm':
                if extension in ['webm']:
                    fmt = "WebM"
                else:
                    fmt = "Matroska"
            elif format_name == 'rm':
                if extension in ['rmvb']:
                    fmt = "RealMedia Variable Bitrate (RMVB)"
                else:
                    fmt = "RealMedia"
        else:
            fmt = extension.upper()

        self.__dp("left StoryBoard._get_format")
        return fmt

    def _get_size(self):
        """Get size of the video file.

        Returns
        -------
        size : int
            Size in bytes.
        size_text: str
            Size as a human readable string, e.g., ``'128MiB'``.

        """

        self.__dp("entered StoryBoard._get_size")
        size = int(self._ffprobe['format']['size'])
        size_text = util.humansize(size)
        self.__dp("left StoryBoard._get_size")
        return (size, size_text)

    def _get_duration(self):
        """Get duration of the video.

        Returns
        -------
        duration : float
            Duration in seconds. ``None`` if duration is not available.
        duration_text : str
            Duration as a human readable string, e.g.,
            ``'00:02:53.33'``. ``None`` if duration is not available.

        """

        self.__dp("entered StoryBoard._get_duration")
        if 'duration' in self._ffprobe['format']:
            duration = float(self._ffprobe['format']['duration'])
            duration_text = util.humantime(duration)
            self.__dp("left StoryBoard._get_duration")
            return (duration, duration_text)
        else:
            self.__dp("left StoryBoard._get_duration")
            return (None, None)

    _SHA_CHUNK_SIZE = 65536
    """Chunk size used when computing the SHA-1 digest."""

    def _get_sha1sum(self, print_progress=False):
        """Get SHA-1 hex digest of the video file.

        In addition to returned the digest, it is also stored in the
        `sha1sum` attribute for future requests. Be aware that computing
        SHA-1 digest is an expensive operation.

        Parameters
        ----------
        print_progress : bool
            Whether to print progress information (to stderr). Default
            is False.

        Returns
        -------
        sha1sum : str
            A SHA-1 hex digest.

        """

        self.__dp("entered StoryBoard._get_sha1sum")
        # directly return if already computed
        if self.sha1sum is not None:
            self.__dp("left StoryBoard._get_sha1sum")
            return self.sha1sum

        if print_progress:
            sys.stderr.write("Computing SHA-1 digest...\n")
        with open(self.path, 'rb') as video:
            sha1 = hashlib.sha1()
            totalsize = os.path.getsize(self.path)
            chunksize = self._SHA_CHUNK_SIZE

            if print_progress:
                pbar = util.ProgressBar(totalsize)
            for chunk in iter(lambda: video.read(chunksize), b''):
                sha1.update(chunk)
                if print_progress:
                    pbar.update(chunksize)
            if print_progress:
                pbar.finish()

            self.sha1sum = sha1.hexdigest()
            self.__dp("left StoryBoard._get_sha1sum")
            return self.sha1sum

    def _get_scan_type(self, ffprobe_bin, print_progress=False):
        """Determine the scan type of the video.

        Parameters
        ----------
        ffprobe_bin : str
            Name/path of the ffprobe binary (should be callable).
        print_progress : bool
            Whether to print progress information (to stderr). Default
            is False.

        Returns
        -------
        scan_type : str
            ``'Progressive scan'``, ``'Interlaced scan'``, or
            ``'Telecined video'``. ``None`` if the file is a pure audio
            file (possibly with album art) or too short (with fewer than
            forty frames).

        Notes
        -----
        In order to determine the scan type, we examie the first forty
        video frames with ffprobe (-show_frames). Each ffprobe frame
        object contains a key named ``interlaced``, which is 0 if the
        frame is progressive or 1 if the frame is interlaced.

        If less than forty video frames are available, then either we
        are dealing with an audio file, or the video file is just too
        short. Either case we do not try to determine the scan type,
        and just set it to None.

        Otherwise, we drop the first twenty frames (since there are
        sometimes junk frames at the beginning), and count the number
        of interlaced frames in the latter twenty frames. If they are
        all progressive or all interlaced, then the answer is
        obvious. If there are 8 interlaced frames out of 20, then it
        is highly probable that the video is telecined. Other than
        that it's pretty confusing, and I would just call it
        interlaced, since a deinterlacer might come in handy anyway.

        Note that this solution assumes that the output format of the
        relevant ffprobe command is

        ::

            {
                "frames": [
                    { frame object ... },
                    { frame object ... },
                    ...

        See https://github.com/zmwangx/storyboard/issues/11 for details.

        """

        # pylint: disable=too-many-branches

        self.__dp("entered StoryBoard._get_scan_type")
        if print_progress:
            sys.stderr.write("Trying to determine scan type...\n")

        ffprobe_args = [
            ffprobe_bin,
            '-select_streams', 'v',
            '-show_frames',
            '-print_format', 'json',
            self.path,
        ]
        proc = subprocess.Popen(ffprobe_args,
                                stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        lines = iter(proc.stdout.readline, b'')

        # skip two lines:
        # {
        #     "frames" : [
        self.__dp(next(lines).decode('utf-8'), newline=False)
        self.__dp(next(lines).decode('utf-8'), newline=False)

        # empty string for incremental storage of json object
        obj_str = ''
        objs = []
        counter = 0
        for line in lines:
            self.__dp(line.decode('utf-8'), newline=False)
            obj_str += line.decode('utf-8').strip()
            try:
                # a complete frame object might be followed by a comma
                # in the array
                if obj_str[-1] == ',':
                    objs.append(json.loads(obj_str[0:-1]))
                else:
                    objs.append(json.loads(obj_str))

                counter += 1
                if print_progress:
                    sys.stderr.write("\rInspecting frame %d/40..." % counter)

                obj_str = ''
                if len(objs) >= 40:
                    proc.terminate()
                    proc.communicate()
                    break
            except ValueError:
                # incomplete frame object
                pass
        if print_progress:
            sys.stderr.write("\n")
        if len(objs) < 40:
            # frame count less than 40, either file is audio or file is
            # video but too short
            self.__dp("left StoryBoard._get_scan_type")
            return None

        # drop the first half of the frame objects
        frames = objs[20:]
        # count interlaced frames in the remaining 20 frames
        num_interlaced = 0
        for frame in frames:
            if 'interlaced_frame' in frame:
                num_interlaced += frame['interlaced_frame']

        self.__dp("left StoryBoard._get_scan_type")
        if num_interlaced == 0:
            return "Progressive scan"
        elif num_interlaced == 20:
            return "Interlaced scan"
        elif num_interlaced == 8:
            # telecined, 3:2 pull down
            return "Telecined video"
        else:
            # confused, see https://github.com/zmwangx/storyboard/issues/11
            return "Interlaced scan"

    def _process_streams(self):
        """Extract per-stream metadata of all streams in the video.

        Extracted metadata are saved to the `streams` attribute.

        """
        self.__dp("entered StoryBoard._process_streams")
        self.streams = []
        for stream in self._ffprobe['streams']:
            self.streams.append(self._process_stream(stream))
        self.__dp("left StoryBoard._process_streams")

    def _process_stream(self, stream_dict):
        """Process a single stream object returned by FFprobe.

        A FFprobe-generated JSON stream object (one single stream) is
        interpreted and save to our in-house Stream object.

        Parameters
        ----------
        stream_dict : dict
            A dict representing a FFprobe-generated JSON stream object
            (one single stream).

        Returns
        -------
        stream : Stream
            A Stream object containing parsed metadata.

        Notes
        -----
        Different codecs are dealt with differently
        here. `_process_video_stream`, `_process_audio_stream` and
        `_process_subtitle_stream` encompass a growing list of codecs I
        frequently encounter. I do not intend to be exhaustive, but
        everyone is welcome to contribute code for their favorite (or
        hated) codecs.

        """

        self.__dp("entered StoryBoard._process_stream")
        if 'codec_type' not in stream_dict:
            stream = Stream()
            stream.type = 'unknown'
            stream.info_string = "Data"
        else:
            codec_type = stream_dict['codec_type']
            if codec_type == "video":
                stream = self._process_video_stream(stream_dict)
            elif codec_type == "audio":
                stream = self._process_audio_stream(stream_dict)
            elif codec_type == "subtitle":
                stream = self._process_subtitle_stream(stream_dict)
            else:
                stream = Stream()
                stream.type = codec_type
                stream.info_string = 'Data'

        stream.index = stream_dict['index']

        self.__dp("left StoryBoard._process_stream")
        return stream

    def _process_video_stream(self, stream_dict):
        """Process video stream object returned by FFprobe.

        Also set Video's `dimension`, `dimension_text`, `dar`,
        `dar_text`, `frame_rate`, and `frame_rate_text` attributes
        (attributes of ``self`` ) if they are available in this video
        stream and they are not already set.

        Parameters
        ----------
        stream_dict : dict
            A dict representing a FFprobe-generated JSON video stream
            object.

        Returns
        -------
        stream : Stream
            A Stream object containing parsed metadata.

        """

        # pylint: disable=too-many-statements,too-many-branches

        self.__dp("entered StoryBoard._process_video_stream")
        sdict = stream_dict  # alias to the long long name

        if sdict['codec_type'] != "video":
            raise ValueError("stream_dict is not a video stream")

        # pylint: disable=invalid-name
        # the stream appears way to often
        s = Stream()
        s.type = "video"

        # codec
        if 'codec_name' not in sdict:
            s.codec = "unknown codec"
        elif sdict['codec_name'] in _VCODEC_MAP:
            codec_name = sdict['codec_name']
            s.codec = _VCODEC_MAP[codec_name]

            # some codecs that need special treatment
            if ((codec_name in {'h264', 'hevc'} and
                 'profile' in sdict and 'level' in sdict)):
                # H.264 and HEVC have profiles and levels. Examples:
                #     H.264 (High Profile level 1.1)
                #     HEVC (Main Profile level 3.0)
                s.codec = ("%s (%s Profile level %.1f)" %
                           (s.codec, sdict['profile'], sdict['level'] / 10.0))
            elif (codec_name in {'mpeg1video', 'mpeg2video'} and
                  'profile' in sdict):
                # MPEG-1 Part 2 and MPEG-2 Part 2 has profiles. Example:
                #     MPEG-1 Part 2 (Main Profile)
                #     MPEG-2 Part 2 (Main Profile)
                s.codec = "%s (%s Profile)" % (s.codec, sdict['profile'])
            elif codec_name in {'mpeg4'} and 'profile' in sdict:
                # MPEG-4 Part 2 has profiles. Example:
                #     MPEG-4 Part 2 (Simple Profile)
                # This case is different in that FFprobe's profile field
                # already includes the word "Profile".
                s.codec = "%s (%s)" % (s.codec, sdict['profile'])
        else:
            s.codec = sdict['codec_long_name']

        # dimension
        s.width = sdict['width']
        s.height = sdict['height']
        s.dimension = (s.width, s.height)
        s.dimension_text = "%dx%d" % (s.width, s.height)
        if self.dimension is None:
            # set video dimension to dimension of the first video stream
            self.dimension = s.dimension
            self.dimension_text = s.dimension_text

        # display aspect ratio (DAR)
        if 'display_aspect_ratio' in sdict:
            s.dar = util.evaluate_ratio(sdict['display_aspect_ratio'])
        if s.dar is not None:
            s.dar_text = sdict['display_aspect_ratio']
        else:
            gcd = fractions.gcd(s.width, s.height)
            reduced_width = s.width // gcd
            reduced_height = s.height // gcd
            s.dar = reduced_width / reduced_height
            s.dar_text = "%d:%d" % (reduced_width, reduced_height)
        if self.dar is None:
            # set video DAR to DAR of the first video stream
            self.dar = s.dar
            self.dar_text = s.dar_text

        # frame rate
        if 'r_frame_rate' in sdict:
            s.frame_rate = util.evaluate_ratio(sdict['r_frame_rate'])
        elif 'avg_frame_rate' in sdict:
            s.frame_rate = util.evaluate_ratio(sdict['avg_frame_rate'])
        else:
            s.frame_rate = None

        if s.frame_rate is not None:
            fps = s.frame_rate
            if abs(fps - int(fps)) < 0.0001:  # integer
                s.frame_rate_text = '%d fps' % int(fps)
            else:
                s.frame_rate_text = "%.2f fps" % fps
        else:
            s.frame_rate_text = None

        if self.frame_rate is None:
            # set video frame rate to that of the first video stream
            self.frame_rate = s.frame_rate
            self.frame_rate_text = s.frame_rate_text

        # bit rate
        if 'bit_rate' in sdict:
            s.bit_rate = float(sdict['bit_rate'])
            s.bit_rate_text = '%d kb/s' % int(round(s.bit_rate / 1000))
        else:
            s.bit_rate = None
            s.bit_rate_text = None

        # assemble info string
        s.info_string = ("Video, %s, %s (DAR %s)" %
                         (s.codec, s.dimension_text, s.dar_text))
        if s.frame_rate_text:
            s.info_string += ", " + s.frame_rate_text
        if s.bit_rate_text:
            s.info_string += ", " + s.bit_rate_text

        self.__dp("left StoryBoard._process_video_stream")
        return s

    def _process_audio_stream(self, stream_dict):
        """Process audio stream object returned by FFprobe.

        Parameters
        ----------
        stream_dict : dict
            A dict representing a FFprobe-generated JSON audio stream
            object.

        Returns
        -------
        stream : Stream
            A Stream object containing parsed metadata.

        """

        # pylint: disable=too-many-statements,too-many-branches

        self.__dp("entered StoryBoard._process_audio_stream")
        sdict = stream_dict  # alias to the long long name

        if sdict['codec_type'] != "audio":
            raise ValueError("stream_dict is not a audio stream")

        # pylint: disable=invalid-name
        # the stream appears way to often
        s = Stream()
        s.type = "audio"

        # codec
        if 'codec_name' not in sdict:
            s.codec = "unknown codec"
        elif sdict['codec_name'] in _ACODEC_MAP:
            codec_name = sdict['codec_name']
            s.codec = _ACODEC_MAP[codec_name]

            # some codecs that need special treatment
            if codec_name == "aac" and 'profile' in sdict:
                if sdict['profile'] == 'LC':
                    profile = 'Low-Complexity'
                elif sdict['profile'] == 'HE-AACv2':
                    profile = 'HE-AAC v2'
                else:
                    profile = sdict['profile']
                s.codec = "AAC (%s)" % profile
        else:
            s.codec = sdict['codec_long_name']

        # bit rate
        if 'bit_rate' in sdict:
            s.bit_rate = float(sdict['bit_rate'])
            s.bit_rate_text = '%d kb/s' % int(round(s.bit_rate / 1000))
        else:
            s.bit_rate = None
            s.bit_rate_text = None

        # language
        if 'tags' in sdict:
            if 'language' in sdict['tags']:
                s.language_code = sdict['tags']['language']
            elif 'LANGUAGE' in sdict['tags']:
                s.language_code = sdict['tags']['LANGUAGE']

        # assemble info string
        if s.language_code:
            s.info_string = "Audio (%s), %s" % (s.language_code, s.codec)
        else:
            s.info_string = "Audio, %s" % s.codec
        if s.bit_rate_text:
            s.info_string += ", " + s.bit_rate_text

        self.__dp("left StoryBoard._process_audio_stream")
        return s

    def _process_subtitle_stream(self, stream_dict):
        """Process subtitle stream object returned by FFprobe.

        Parameters
        ----------
        stream_dict : dict
            A dict representing a FFprobe-generated JSON subtitle stream
            object.

        Returns
        -------
        stream : Stream
            A Stream object containing parsed metadata.

        """

        # pylint: disable=too-many-statements,too-many-branches

        self.__dp("entered StoryBoard._process_subtitle_stream")
        sdict = stream_dict  # alias to the long long name

        if sdict['codec_type'] != "subtitle":
            raise ValueError("stream_dict is not a subtitle stream")

        # pylint: disable=invalid-name
        # the stream appears way to often
        s = Stream()
        s.type = "subtitle"

        if 'codec_name' not in sdict:
            if (('codec_tag_string' in sdict and
                 sdict['codec_tag_string'] == 'c608')):
                s.codec = 'EIA-608'
            else:
                s.codec = "unknown codec"
        elif sdict['codec_name'] in _SCODEC_MAP:
            codec_name = sdict['codec_name']
            s.codec = _SCODEC_MAP[codec_name]
        else:
            s.codec = sdict['codec_long_name']

        # language
        if 'tags' in sdict:
            if 'language' in sdict['tags']:
                s.language_code = sdict['tags']['language']
            elif 'LANGUAGE' in sdict['tags']:
                s.language_code = sdict['tags']['LANGUAGE']

        # assemble info string
        if s.language_code:
            s.info_string = "Subtitle (%s), %s" % (s.language_code, s.codec)
        else:
            s.info_string = "Subtitle, %s" % s.codec

        self.__dp("left StoryBoard._process_subtitle_stream")
        return s

    def __dp(self, string, newline=True):
        """Debug printing."""
        if hasattr(self, "_Video__debug") and self.__debug:
            if newline:
                print(string, file=sys.stderr)
            else:
                sys.stderr.write(string)
            sys.stderr.flush()


def main():
    """CLI interface."""

    # pylint: disable=too-many-locals,too-many-statements,too-many-branches

    description = """Print video metadata.

    You may supply a list of videos, and the output for each video will
    be followed by a blank line to distinguish it from others. Below is
    the list of available options and their brief explanations. Some of
    the options can also be stored in a configuration file,
    $XDG_CONFIG_HOME/storyboard/storyboard.conf (or if $XDG_CONFIG_HOME
    is not defined, ~/.config/storyboard/storyboard.conf), under the
    "metadata-cli" section (the conf file format follows that described
    in https://docs.python.org/3/library/configparser.html).

    For more detailed explanations, see
    http://storyboard.rtfd.org/en/stable/metadata-cli.html (or replace
    "stable" with the version you are using).
    """
    parser = argparse.ArgumentParser(description=description)
    parser.add_argument(
        'videos', nargs='+', metavar='VIDEO',
        help="Path(s) to the video file(s).")
    parser.add_argument(
        '--ffprobe-bin', metavar='NAME',
        help="""The name/path of the ffprobe binary. The binay is
        guessed from OS type if this option is not specified.""")
    parser.add_argument(
        '--include-sha1sum', '-s', action='store_const', const=True,
        help="Include SHA-1 digest of the video(s).")
    parser.add_argument(
        '--exclude-sha1sum', action='store_true',
        help="""Exclude SHA-1 digest of the video(s). Overrides
        '--include-sha1sum'. This option is only useful if
        include_sha1sum is turned on by default in the config file.""")
    parser.add_argument(
        '--verbose', '-v', choices=['auto', 'on', 'off'],
        nargs='?', const='auto',
        help="""Whether to print progress information to stderr. Default
        is 'auto'.""")
    parser.add_argument(
        '--version', action='version', version=version.__version__)
    cli_args = parser.parse_args()

    if 'XDG_CONFIG_HOME' in os.environ:
        config_file = os.path.join(os.environ['XDG_CONFIG_HOME'],
                                   'storyboard/storyboard.conf')
    else:
        config_file = os.path.expanduser('~/.config/storyboard/storyboard.conf')

    defaults = {
        'ffprobe_bin': fflocate.guess_bins()[1],
        'include_sha1sum': False,
        'verbose': 'auto',
    }

    optreader = util.OptionReader(
        cli_args=cli_args,
        config_files=config_file,
        section='metadata-cli',
        defaults=defaults,
    )
    ffprobe_bin = optreader.opt('ffprobe_bin')
    include_sha1sum = optreader.opt('include_sha1sum', opttype=bool)
    if cli_args.exclude_sha1sum:
        # force override
        include_sha1sum = False
    verbose = optreader.opt('verbose')
    if verbose == 'on':
        print_progress = True
    elif verbose == 'off':
        print_progress = False
    else:
        if verbose != 'auto':
            msg = ("warning: '%s' is a not a valid argument to --verbose; "
                   "ignoring and using 'auto' instead\n" % verbose)
            sys.stderr.write(msg)
        if include_sha1sum and sys.stderr.isatty():
            print_progress = True
        else:
            print_progress = False

    # test ffprobe_bin
    try:
        fflocate.check_bins((None, ffprobe_bin))
    except OSError:
        msg = ("fatal error: '%s' does not exist on PATH or is corrupted "
               "(expected FFprobe)\n" % ffprobe_bin)
        sys.stderr.write(msg)
        exit(1)

    # real stuff happens from here
    returncode = 0
    for video in cli_args.videos:
        # pylint: disable=invalid-name
        try:
            v = Video(video, params={
                'ffprobe_bin': ffprobe_bin,
                'print_progress': print_progress,
            })
        except OSError as err:
            sys.stderr.write("error: %s\n\n" % str(err))
            returncode = 1
            continue

        metadata_string = v.format_metadata(params={
            'include_sha1sum': include_sha1sum,
            'print_progress': print_progress,
        })

        if print_progress:
            # print one empty line to separate progress info and output
            # content
            sys.stderr.write("\n")
        print(metadata_string)
        print('')
    return returncode


if __name__ == "__main__":
    exit(main())



#!/usr/bin/env python3

"""Check existence of ffmpeg and ffprobe.

Routines
--------
.. autosummary::
    guess_bins
    check_bins

----

"""

import os
import subprocess


def guess_bins():
    """Guess ffmpeg and ffprobe binary names based on OS.

    Returns
    -------
    bins : tuple
        A tuple ``(ffmpeg_bin, ffprobe_bin)`` of two strings, where
        `ffmpeg_bin` is the guessed ffmpeg binary, and `ffprobe_bin` is
        the guessed ffprobe binary.

    """

    if os.name == 'nt':
        return ('ffmpeg.exe', 'ffprobe.exe')
    else:
        return ('ffmpeg', 'ffprobe')


def check_bins(bins):
    """Check existance of ffmpeg and ffprobe binaries.

    Parameters
    ----------
    bins : tuple
        A tuple `(ffmpeg_bin, ffprobe_bin)`` of the binary
        names/paths. Either of the two can be ``None``, in which case
        the corresponding binary is not checked.

    Returns
    -------
    True
        If check is successful.

    Raises
    ------
    OSError
        If check fails.

    """

    with open(os.devnull, 'wb') as devnull:
        for binary in bins:
            if binary is None:
                continue
            try:
                subprocess.check_call([binary, '-version'],
                                      stdout=devnull, stderr=devnull)
            except subprocess.CalledProcessError:
                raise OSError("%s may be corrupted" % binary)
            except OSError:
                raise OSError("%s not found on PATH" % binary)
    return True



#!/usr/bin/env python3

"""Generate video storyboards with metadata reports.

Classes
-------
.. autosummary::
    Font
    StoryBoard

Routines
--------
.. autosummary::
    create_thumbnail
    draw_text_block
    tile_images
    main

----

"""

# pylint: disable=too-many-lines

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import argparse
import pkg_resources
import os
import sys
import tempfile

from PIL import Image, ImageDraw, ImageFont

from storyboard import fflocate
from storyboard.frame import extract_frame as _extract_frame
from storyboard import metadata
from storyboard import util
from storyboard.util import read_param as _read_param
from storyboard import version


# load default font
DEFAULT_FONT_FILE = pkg_resources.resource_filename(
    __name__,
    'SourceCodePro-Regular.otf'
)
DEFAULT_FONT_SIZE = 16


# pylint: disable=too-many-locals,invalid-name
# In this file we use a lot of short local variable names to save space.
# These short variable names are carefully documented when not obvious.


class Font(object):

    """Wrapper for a font loaded by PIL.ImageFont.

    Parameters
    ----------
    font_file : str
        Path to the font file to be loaded. If ``None``, load the
        default font defined by the module variable
        ``DEFAULT_FONT_FILE``. Default is ``None``. Note that the font
        must be supported by FreeType.
    font_size : int
        Font size to be loaded. If ``None``, use the default font size
        defined by the module variable ``DEFAULT_FONT_SIZE``. Default is
        ``None``.

    Raises
    ------
    OSError
        If the font cannot be loaded (by ``PIL.ImageFont.truetype``).

    Attributes
    ----------
    obj
        A Pillow font object, e.g., of the type
        ``PIL.ImageFont.FreeTypeFont``.
    size : int
        The font size.

    Notes
    -----
    We are creating this wrapper because there's no guarantee that a
    font loaded by Pillow will have the ``.size`` property, and
    sometimes it certainly doesn't (try
    ``PIL.ImageFont.load_default()``, for instance). The font size,
    however, is crucial for some of other drawings, so we would like to
    keep it around all the time.

    """

    # pylint: disable=too-few-public-methods

    def __init__(self, font_file=None, font_size=None):
        """Initialize the Font class.

        See class docstring for parameters of the constructor.

        """

        if font_file is None:
            font_file = DEFAULT_FONT_FILE
        if font_size is None:
            font_size = DEFAULT_FONT_SIZE

        try:
            self.obj = ImageFont.truetype(font_file, size=font_size)
        except IOError:
            raise OSError("font file '%s' cannot be loaded" % font_file)
        self.size = font_size


def draw_text_block(canvas, xy, text, params=None):
    """Draw a block of text.

    You need to specify a canvas to draw upon. If you are not sure about
    the size of the canvas, there is a `dry_run` option (see "Other
    Parameters" that help determine the size of the text block before
    creating the canvas.

    Parameters
    ----------
    canvas : PIL.ImageDraw.Image
        The canvas to draw the text block upon. If the `dry_run` option
        is on (see "Other Parameters"), `canvas` can be ``None``.
    xy : tuple
        Tuple ``(x, y)`` consisting of x and y coordinates of the
        topleft corner of the text block.
    text : str
        Text to be drawn, can be multiline.
    params : dict, optional
        Optional parameters enclosed in a dict. Default is ``None``.
        See the "Other Parameters" section for understood key/value
        pairs.

    Returns
    -------
    (width, height)
        Size of text block.

    Other Parameters
    ----------------
    font : Font, optional
        Default is the font constructed by ``Font()`` without arguments.
    color : color, optional
        Color of text; can be in any color format accepted by Pillow
        (used for the ``fill`` argument of
        ``PIL.ImageDraw.Draw.text``). Default is ``'black'``.
    spacing : float, optional
        Line spacing as a float. Default is 1.2.
    dry_run : bool, optional
        If ``True``, do not draw anything, only return the size of the
        text block. Default is ``False``.

    """

    if params is None:
        params = {}
    x, y = xy
    font = _read_param(params, 'font', Font())
    color = _read_param(params, 'color', 'black')
    spacing = _read_param(params, 'spacing', 1.2)
    dry_run = _read_param(params, 'dry_run', False)

    if not dry_run:
        draw = ImageDraw.Draw(canvas)

    line_height = int(round(font.size * spacing))
    width = 0
    height = 0
    for line in text.splitlines():
        w, _ = font.obj.getsize(line)
        if not dry_run:
            draw.text((x, y), line, fill=color, font=font.obj)
        if w > width:
            width = w  # update width to that of the current widest line
        height += line_height
        y += line_height
    return (width, height)


def create_thumbnail(frame, width, params=None):
    """Create thumbnail of a video frame.

    The timestamp of the frame can be overlayed to the thumbnail. See
    "Other Parameters" to how to enable this feature and relevant
    options.

    Parameters
    ----------
    frame : storyboard.frame.Frame
        The video frame to be thumbnailed.
    width : int
        Width of the thumbnail.
    params : dict, optional
        Optional parameters enclosed in a dict. Default is ``None``.
        See the "Other Parameters" section for understood key/value
        pairs.

    Returns
    -------
    thumbnail : PIL.Image.Image

    Other Parameters
    ----------------
    aspect_ratio : float, optional
        Aspect ratio of the thumbnail. If ``None``, use the aspect ratio
        (only considering the pixel dimensions) of the frame. Default is
        ``None``.
    draw_timestamp : bool, optional
        Whether to draw frame timestamp over the timestamp. Default is
        ``False``.
    timestamp_font : Font, optional
        Font for the timestamp, if `draw_timestamp` is ``True``.
        Default is the font constructed by ``Font()`` without arguments.
    timestamp_align : {'right', 'center', 'left'}, optional
        Horizontal alignment of the timestamp over the thumbnail, if
        `draw_timestamp` is ``True``. Default is ``'right'``. Note that
        the timestamp is always vertically aligned towards the bottom of
        the thumbnail.

    """

    if params is None:
        params = {}
    if 'aspect_ratio' in params:
        aspect_ratio = params['aspect_ratio']
    else:
        image_width, image_height = frame.image.size
        aspect_ratio = image_width / image_height
    height = int(round(width / aspect_ratio))
    size = (width, height)
    draw_timestamp = _read_param(params, 'draw_timestamp', False)
    if draw_timestamp:
        timestamp_font = _read_param(params, 'timestamp_font', Font())
        timestamp_align = _read_param(params, 'timestamp_align', 'right')

    thumbnail = frame.image.resize(size, Image.LANCZOS)

    if draw_timestamp:
        draw = ImageDraw.Draw(thumbnail)

        timestamp_text = util.humantime(frame.timestamp, ndigits=0)
        timestamp_width, timestamp_height = \
            draw.textsize(timestamp_text, timestamp_font.obj)

        # calculate upperleft corner of the timestamp overlay
        # we hard code a margin of 5 pixels
        timestamp_y = height - 5 - timestamp_height
        if timestamp_align == 'right':
            timestamp_x = width - 5 - timestamp_width
        elif timestamp_align == 'left':
            timestamp_x = 5
        elif timestamp_align == 'center':
            timestamp_x = int((width - timestamp_width) / 2)
        else:
            raise ValueError("timestamp alignment option '%s' not recognized"
                             % timestamp_align)

        # draw white timestamp with 1px thick black border
        for x_offset in range(-1, 2):
            for y_offset in range(-1, 2):
                draw.text((timestamp_x + x_offset, timestamp_y + y_offset),
                          timestamp_text,
                          fill='black', font=timestamp_font.obj)
        draw.text((timestamp_x, timestamp_y),
                  timestamp_text,
                  fill='white', font=timestamp_font.obj)

    return thumbnail


def tile_images(images, tile, params=None):
    """
    Combine images into a composite image through 2D tiling.

    For example, 16 thumbnails can be combined into an 4x4 array. As
    another example, three images of the same width (think of the
    metadata sheet, the bare storyboard, and the promotional banner) can
    be combined into a 1x3 array, i.e., assembled vertically.

    The image list is processed column-first.

    Note that except if you use the `tile_size` option (see "Other
    Parameters"), you should make sure that images passed into this
    function satisfy the condition that the widths of all images in each
    column and the heights of all images in each row match perfectly;
    otherwise, this function will give up and raise a ValueError.

    Parameters
    ----------
    images : list
        A list of PIL.Image.Image objects, satisfying the necessary
        height and width conditions (see explanation above).
    tile : tuple
        A tuple ``(m, n)`` indicating `m` columns and `n` rows. The
        product of `m` and `n` should be the length of `images`, or a
        `ValueError` will arise.
    params : dict, optional
        Optional parameters enclosed in a dict. Default is ``None``.
        See the "Other Parameters" section for understood key/value
        pairs.

    Returns
    -------
    PIL.Image.Image
        The composite image.

    Raises
    ------
    ValueError
        If the length of `images` does not match the product of columns
        and rows (as defined in `tile`), or the widths and heights of
        the images do not satisfy the necessary consistency conditions.

    Other Parameters
    ----------------
    tile_size : tuple, optional
        A tuple ``(width, height)``. If this parameter is specified,
        width and height consistency conditions won't be checked, and
        every image will be resized to (width, height) when
        combined. Default is ``None``.
    tile_spacing : tuple, optional
        A tuple ``(hor, ver)`` specifying the horizontal and vertical
        spacing between adjacent tiles. Default is ``(0, 0)``.
    margins : tuple, optional
        A tuple ``(hor, ver)`` specifying the horizontal and vertical
        margins (padding) around the combined image. Default is ``(0,
        0)``.
    canvas_color : color, optional
        A color in any format recognized by Pillow. This is only
        relevant if you have nonzero tile spacing or margins, when the
        background shines through the spacing or margins. Default is
        ``'white'``.
    close_separate_images : bool
        Whether to close the separate after combining. Closing the
        images will release the corresponding resources. Default is
        ``False``.

    """

    # pylint: disable=too-many-branches

    if params is None:
        params = {}
    cols, rows = tile
    if len(images) != cols * rows:
        msg = "{} images cannot fit into a {}x{}={} array".format(
            len(images), cols, rows, cols * rows)
        raise ValueError(msg)
    hor_spacing, ver_spacing = _read_param(params, 'tile_spacing', (0, 0))
    hor_margin, ver_margin = _read_param(params, 'margins', (0, 0))
    canvas_color = _read_param(params, 'canvas_color', 'white')
    close_separate_images = _read_param(params, 'close_separate_images', False)
    if 'tile_size' in params and params['tile_size'] is not None:
        tile_size = params['tile_size']
        tile_width, tile_height = tile_size
        canvas_width = (tile_width * cols + hor_spacing * (cols - 1) +
                        hor_margin * 2)
        canvas_height = (tile_height * rows + ver_spacing * (rows - 1) +
                         ver_margin * 2)
        resize = True
    else:
        # check column width consistency, bark if not
        # calculate total width along the way
        canvas_width = hor_spacing * (cols - 1) + hor_margin * 2
        for col in range(0, cols):
            # reference width set by the first image in the column
            ref_index = 0 * cols + col
            ref_width, ref_height = images[ref_index].size
            canvas_width += ref_width
            for row in range(1, rows):
                index = row * cols + col
                width, height = images[index].size
                if width != ref_width:
                    msg = ("the width of image #{} "
                           "(row #{}, col #{}, {}x{}) "
                           "does not agree with that of image #{} "
                           "(row #{}, col #{}, {}x{})".format(
                               index, row, col, width, height,
                               ref_index, 0, col, ref_width, ref_height,
                           ))
                    raise ValueError(msg)
        # check row height consistency, bark if not
        # calculate total height along the way
        canvas_height = ver_spacing * (rows - 1) + ver_margin * 2
        for row in range(0, rows):
            # reference width set by the first image in the column
            ref_index = row * cols + 0
            ref_width, ref_height = images[ref_index].size
            canvas_height += ref_height
            for col in range(1, cols):
                index = row * cols + col
                width, height = images[index].size
                if height != ref_height:
                    msg = ("the height of image #{} "
                           "(row #{}, col #{}, {}x{}) "
                           "does not agree with that of image #{} "
                           "(row #{}, col #{}, {}x{})".format(
                               index, row, col, width, height,
                               ref_index, 0, col, ref_width, ref_height,
                           ))
                    raise ValueError(msg)
        # passed tests, will assemble as is
        resize = False

    # start assembling images
    canvas = Image.new('RGBA', (canvas_width, canvas_height), canvas_color)
    y = ver_margin
    for row in range(0, rows):
        x = hor_margin
        for col in range(0, cols):
            image = images[row * cols + col]

            if resize:
                canvas.paste(image.resize(tile_size, Image.LANCZOS), (x, y))
            else:
                canvas.paste(image, (x, y))

            # accumulate width of this column, as well as horizontal spacing
            x += image.size[0] + hor_spacing
        # accumulate height of this row, as well as vertical spacing
        y += images[row * cols + 0].size[1] + ver_spacing

    if close_separate_images:
        for image in images:
            image.close()

    return canvas


class StoryBoard(object):
    """Class for creating video storyboard.

    Parameters
    ----------
    video
        Either a string specifying the path to the video file, or a
        ``storyboard.metadata.Video`` object.
    params : dict, optional
        Optional parameters enclosed in a dict. Default is
        ``None``. See the "Other Parameters" section for understood
        key/value pairs.

    Raises
    ------
    OSError
        If ffmpeg and ffprobe binaries do not exist or seem corrupted,
        or if the video does not exist or cannot be recognized by
        FFprobe.

    Other Parameters
    ----------------
    bins : tuple, optional
        A tuple (ffmpeg_bin, ffprobe_bin), specifying the name of path
        of FFmpeg and FFprobe's binaries on your system. If ``None``,
        the bins are guessed according to type of OS, using
        ``storyboard.fflocate.guess_bins`` (e.g., on OS X and Linux
        systems, the natural names are ``'ffmpeg'`` and ``'ffprobe'``;
        on Windows, the names have ``'.exe'`` suffixes). Default is
        ``None``.
    frame_codec : str, optional
        Image codec to use when extracting frames using FFmpeg. Default
        is ``'png'``. Use this option with caution only if your FFmpeg
        cannot encode PNG, which is unlikely.
    video_duration : float, optional
        Duration of the video in seconds, passed to the
        ``storyboard.metadata.Video`` constructor. If ``None``, extract
        the duration from video container metadata. Default is
        ``None``. You should rarely need this option, unless the
        duration of the video cannot be extracted from video container
        metadata, or the duration extracted is wrong. Either case is
        fatal to the storyboard (since the frames extracted depend on
        the duration), and this option provides a fallback. See `#3
        <https://github.com/zmwangx/storyboard/issues/3>`_ for details.
    print_progress : bool, optional
        Whether to print progress information (to stderr). Default is
        ``False``.

    Attributes
    ----------
    video : storyboard.metadata.Video
    frames : list
        List of equally spaced frames in the video, as
        ``storyboard.frame.Frame`` objects. The list is empty after
        `__init__`. See the `gen_frames` method.

    Notes
    -----
    For developers: there are two private attributes. ``_bins`` is a
    tuple of two strs holding the name or path of the ffmpeg and ffprobe
    binaries; ``_frame_codec`` is a str holding the image codec used by
    FFmpeg when generating frames (usually no one needs to touch this).

    """

    def __init__(self, video, params=None):
        """Initialize the StoryBoard class.

        See the module docstring for parameters and exceptions.

        """

        if params is None:
            params = {}
        if 'bins' in params and params['bins'] is not None:
            bins = params['bins']
            assert isinstance(bins, tuple) and len(bins) == 2
        else:
            bins = fflocate.guess_bins()
        frame_codec = _read_param(params, 'frame_codec', 'png')
        video_duration = _read_param(params, 'video_duration', None)
        print_progress = _read_param(params, 'print_progress', False)

        fflocate.check_bins(bins)

        # seek frame by frame if video duration is specially given
        # (indicating that normal input seeking may not work)
        self._seek_frame_by_frame = video_duration is not None

        self._bins = bins
        if isinstance(video, metadata.Video):
            self.video = video
        elif isinstance(video, str):
            self.video = metadata.Video(video, params={
                'ffprobe_bin': bins[1],
                'video_duration': video_duration,
                'print_progress': print_progress,
            })
        else:
            raise ValueError("expected str or storyboard.metadata.Video "
                             "for the video argument, got %s" %
                             type(video).__name__)
        self.frames = []
        self._frame_codec = frame_codec

    def gen_storyboard(self, params=None):
        """Generate full storyboard.

        A full storyboard has three sections, arranged vertically in the
        order listed: a metadata sheet, a bare storyboard, and a
        promotional banner.

        The metadata sheet consists of formatted metadata generated by
        ``storyboard.metadata.Video.format_metadata``; you may choose
        whether to include the SHA-1 hex digest of the video file (see
        `include_sha1sum` in "Other Parameters"). The bare storyboard is
        an array (usually 4x4) of thumbnails, generated from equally
        spaced frames from the video. The promotional banner briefly
        promotes this package (storyboard).

        The metadata sheet and promotional banner are optional and can
        be turned off individually. See `include_metadata_sheet` and
        `include_promotional_banner` in "Other Parameters".

        `Here <https://i.imgur.com/9T2zM8R.jpg>`_ is a basic example of
        a full storyboard.

        Parameters
        ----------
        params : dict, optional
            Optional parameters enclosed in a dict. Default is
            ``None``. See the "Other Parameters" section for understood
            key/value pairs.

        Returns
        -------
        full_storyboard : PIL.Image.Image

        Other Parameters
        ----------------
        include_metadata_sheet: bool, optional
            Whether to include a video metadata sheet in the
            storyboard. Default is ``True``.
        include_promotional_banner: bool, optional
            Whether to include a short promotional banner about this
            package (storyboard) at the bottom of the
            storyboard. Default is ``True``.

        background_color : color, optional
            Background color of the storyboard, in any color format
            recognized by Pillow. Default is ``'white'``.
        section_spacing : int, optional
            Vertical spacing between adjacent sections (metadata sheet
            and bare storyboard, bare storyboard and promotional
            banner). If ``None``, use the vertical tile spacing (see
            `tile_spacing`). Default is ``None``.
        margins : tuple, optional
            A tuple ``(hor, ver)`` specifying the horizontal and
            vertical margins (padding) around the entire storyboard (all
            sections included). Default is ``(10, 10)``.

        tile : tuple, optional
            A tuple ``(cols, rows)`` specifying the number of columns
            and rows for the array of thumbnails in the
            storyboard. Default is ``(4, 4)``.
        tile_spacing : tuple, optional
            A tuple ``(hor, ver)`` specifying the horizontal and
            vertical spacing between adjacent thumbnails. Default is
            ``(8, 6)``.

        thumbnail_width : int, optional
            Width of each thumbnail. Default is 480 (as in 480x270 for a
            16:9 video).
        thumbnail_aspect_ratio : float, optional
            Aspect ratio of generated thumbnails. If ``None``, first try
            to use the display aspect ratio of the video
            (``self.video.dar``), then the aspect ratio of the frames if
            ``self.video.dar`` is not available. Default is ``None``.

        draw_timestamp : bool, optional
            Whether to draw frame timestamps on top of thumbnails (as an
            overlay).  Default is ``True``.
        timestamp_font : Font, optional
            Font used for timestamps, if `draw_timestamp` is
            ``True``. Default is the font constructed by ``Font()``
            without arguments.
        timestamp_align : {'right', 'center', 'left'}, optional
            Horizontal alignment of timestamps over the thumbnails, if
            `draw_timestamp` is ``True``. Default is ``'right'``. Note
            that timestamps are always vertically aligned to the bottom.

        text_font: Font, optional
            Font used for metadata sheet and promotional banner. Default
            is the font constructed by ``Font()`` without arguments.
        text_color: color, optional
            Color of metadata and promotional text, in any format
            recognized by Pillow. Default is ``'black'``.
        line_spacing: float, optional
            Line spacing of metadata text as a float, e.g., 1.0 for
            single spacing, 1.5 for one-and-a-half spacing, and 2.0 for
            double spacing. Default is 1.2.

        include_sha1sum: bool, optional
            Whether to include the SHA-1 hex digest of the video file in
            the metadata fields. Default is ``False``. Be aware that
            computing SHA-1 digest is an expensive operation.

        print_progress : bool, optional
            Whether to print progress information (to stderr). Default
            is ``False``.

        """

        # process parameters -- a ton of them
        if params is None:
            params = {}
        include_metadata_sheet = _read_param(
            params, 'include_metadata_sheet', True)
        include_promotional_banner = _read_param(
            params, 'include_promotional_banner', True)
        background_color = _read_param(params, 'background_color', 'white')
        margins = _read_param(params, 'margins', (10, 10))
        tile = _read_param(params, 'tile', (4, 4))
        tile_spacing = _read_param(params, 'tile_spacing', (8, 6))
        if (('section_spacing' in params and
             params['section_spacing'] is not None)):
            section_spacing = params['section_spacing']
        else:
            section_spacing = tile_spacing[1]
        thumbnail_width = _read_param(params, 'thumbnail_width', 480)
        thumbnail_aspect_ratio = _read_param(
            params, 'thumbnail_aspect_ratio', None)
        draw_timestamp = _read_param(params, 'draw_timestamp', True)
        timestamp_font = _read_param(params, 'timestamp_font', Font())
        timestamp_align = _read_param(params, 'timestamp_align', 'right')
        text_font = _read_param(params, 'text_font', Font())
        text_color = _read_param(params, 'text_color', 'black')
        line_spacing = _read_param(params, 'line_spacing', 1.2)
        include_sha1sum = _read_param(params, 'include_sha1sum', False)
        print_progress = _read_param(params, 'print_progress', False)

        # draw bare storyboard, metadata sheet, and promotional banner
        if print_progress:
            sys.stderr.write("Generating main storyboard...\n")
        bare_storyboard = self._gen_bare_storyboard(
            tile, thumbnail_width,
            params={
                'tile_spacing': tile_spacing,
                'background_color': background_color,
                'thumbnail_aspect_ratio': thumbnail_aspect_ratio,
                'draw_timestamp': draw_timestamp,
                'timestamp_font': timestamp_font,
                'timestamp_align': timestamp_align,
                'print_progress': print_progress,
            }
        )
        total_width, _ = bare_storyboard.size

        if include_metadata_sheet:
            if print_progress:
                sys.stderr.write("Generating metadata sheet...\n")
            metadata_sheet = self._gen_metadata_sheet(total_width, params={
                'text_font': text_font,
                'text_color': text_color,
                'line_spacing': line_spacing,
                'background_color': background_color,
                'include_sha1sum': include_sha1sum,
                'print_progress': print_progress,
            })

        if include_promotional_banner:
            if print_progress:
                sys.stderr.write("Generating promotional banner...\n")
            banner = self._gen_promotional_banner(total_width, params={
                'text_font': text_font,
                'text_color': text_color,
                'background_color': background_color,
            })

        # combine different sections
        if print_progress:
            sys.stderr.write("Assembling pieces...\n")
        sections = []
        if include_metadata_sheet:
            sections.append(metadata_sheet)
        sections.append(bare_storyboard)
        if include_promotional_banner:
            sections.append(banner)
        storyboard = tile_images(sections, (1, len(sections)), params={
            'tile_spacing': (0, section_spacing),
            'margins': margins,
            'canvas_color': background_color,
            'close_separate_images': True,
        })

        return storyboard

    def gen_frames(self, count, params=None):
        """Extract equally spaced frames from the video.

        When tasked with extracting N frames, this method extracts them
        at positions 1/2N, 3/2N, 5/2N, ... , (2N-1)/2N of the video. The
        extracted frames are stored in the `frames` attribute.

        Note that new frames are extracted only if the number of
        existing frames in the `frames` attribute doesn't match the
        specified `count` (0 at instantiation), in which case new frames
        are extracted to match the specification, and the `frames`
        attribute is overwritten.

        Parameters
        ----------
        count : int
            Number of (equally-spaced) frames to generate.
        params : dict, optional
            Optional parameters enclosed in a dict. Default is
            ``None``. See the "Other Parameters" section for understood
            key/value pairs.

        Returns
        -------
        None
            Access the generated frames through the `frames` attribute.

        RAISES
        ------
        OSError
            If frame extraction with FFmpeg fails.

        Other Parameters
        ----------------
        print_progress : bool, optional
            Whether to print progress information (to stderr). Default
            is False.

        """

        if params is None:
            params = {}
        print_progress = _read_param(params, 'print_progress', False)

        if len(self.frames) == count:
            return

        duration = self.video.duration
        interval = duration / count
        timestamps = [interval * (i + 1/2) for i in range(0, count)]
        counter = 0
        for timestamp in timestamps:
            counter += 1
            if print_progress:
                sys.stderr.write("\rExtracting frame %d/%d..." %
                                 (counter, count))
            try:
                frame = _extract_frame(self.video.path, timestamp, params={
                    'ffmpeg_bin': self._bins[0],
                    'codec': self._frame_codec,
                    'frame_by_frame': self._seek_frame_by_frame,
                })
                self.frames.append(frame)
            except:
                # \rExtracting frame %d/%d... isn't terminated by
                # newline yet
                if print_progress:
                    sys.stderr.write("\n")
                raise
        if print_progress:
            sys.stderr.write("\n")

    def _gen_bare_storyboard(self, tile, thumbnail_width, params=None):
        """Generate bare storyboard (thumbnails only).

        Parameters
        ----------
        tile : tuple
            A tuple ``(cols, rows)`` specifying the number of columns
            and rows for the array of thumbnails.
        thumbnail_width : int
            Width of each thumbnail.
        params : dict, optional
            Optional parameters enclosed in a dict. Default is
            ``None``. See the "Other Parameters" section for understood
            key/value pairs.

        Returns
        -------
        base_storyboard : PIL.Image.Image

        Other Parameters
        ----------------
        tile_spacing : tuple, optional
            See the `tile_spacing` parameter of the `tile_images`
            function. Default is ``(0, 0)``.

        background_color : color, optional
            See the `canvas_color` paramter of the `tile_images`
            function. Default is ``'white'``.

        thumbnail_aspect_ratio : float, optional
            Aspect ratio of generated thumbnails. If ``None``, first try
            to use the display aspect ratio of the video
            (``self.video.dar``), then the aspect ratio of the frames if
            ``self.video.dar`` is not available. Default is ``None``.

        draw_timestamp : bool, optional
            See the `draw_timestamp` parameter of the `create_thumbnail`
            function. Default is ``False``.
        timestamp_font : Font, optional
            See the `timestamp_font` parameter of the `create_thumbnail`
            function. Default is the font constructed by ``Font()``
            without arguments.
        timestamp_align : {'right', 'center', 'left'}, optional
            See the `timestamp_align` parameter of the
            `create_thumbnail` function. Default is ``'right'``.

        print_progress : bool, optional
            Whether to print progress information (to stderr). Default
            is False.

        """

        if params is None:
            params = {}
        tile_spacing = _read_param(params, 'tile_spacing', (0, 0))
        background_color = _read_param(params, 'background_color', 'white')
        if (('thumbnail_aspect_ratio' in params and
             params['thumbnail_aspect_ratio'] is not None)):
            thumbnail_aspect_ratio = params['thumbnail_aspect_ratio']
        elif self.video.dar is not None:
            thumbnail_aspect_ratio = self.video.dar
        else:
            # defer calculation to after generating frames
            thumbnail_aspect_ratio = None
        draw_timestamp = _read_param(params, 'draw_timestamp', False)
        if draw_timestamp:
            timestamp_font = _read_param(params, 'timestamp_font', Font())
            timestamp_align = _read_param(params, 'timestamp_align', 'right')
        print_progress = _read_param(params, 'print_progress', False)

        cols, rows = tile
        if (not(isinstance(cols, int) and isinstance(rows, int) and
                cols > 0 and rows > 0)):
            raise ValueError('tile is not a tuple of positive integers')
        thumbnail_count = cols * rows
        self.gen_frames(cols * rows, params={
            'print_progress': print_progress,
        })
        if thumbnail_aspect_ratio is None:
            frame_size = self.frames[0].image.size
            thumbnail_aspect_ratio = frame_size[0] / frame_size[1]

        thumbnails = []
        counter = 0
        for frame in self.frames:
            counter += 1
            if print_progress:
                sys.stderr.write("\rGenerating thumbnail %d/%d..." %
                                 (counter, thumbnail_count))
            thumbnails.append(create_thumbnail(frame, thumbnail_width, params={
                'aspect_ratio': thumbnail_aspect_ratio,
                'draw_timestamp': draw_timestamp,
                'timestamp_font': timestamp_font,
                'timestamp_align': timestamp_align,
            }))
        if print_progress:
            sys.stderr.write("\n")

        if print_progress:
            sys.stderr.write("Tiling thumbnails...\n")
        return tile_images(thumbnails, tile, params={
            'tile_spacing': tile_spacing,
            'canvas_color': background_color,
            'close_separate_images': True,
        })

    def _gen_metadata_sheet(self, total_width, params=None):
        """Generate metadata sheet.

        Parameters
        ----------
        total_width : int
            Total width of the metadata sheet. Usually determined by the
            width of the bare storyboard.
        params : dict, optional
            Optional parameters enclosed in a dict. Default is
            ``None``. See the "Other Parameters" section for understood
            key/value pairs.

        Returns
        -------
        metadata_sheet : PIL.Image.Image

        Other Parameters
        ----------------
        text_font : Font, optional
            Default is the font constructed by ``Font()`` without
            arguments.
        text_color: color, optional
            Default is 'black'.
        line_spacing : float, optional
            Line spacing as a float. Default is 1.2.
        background_color: color, optional
            Default is 'white'.
        include_sha1sum: bool, optional
            See the `include_sha1sum` option of
            `storyboard.metadata.Video.format_metadata`. Beware that
            computing SHA-1 digest is an expensive operation.
        print_progress : bool, optional
            Whether to print progress information (to stderr). Default
            is False.

        """

        if params is None:
            params = {}
        text_font = _read_param(params, 'text_font', Font())
        text_color = _read_param(params, 'text_color', 'black')
        line_spacing = _read_param(params, 'line_spacing', 1.2)
        background_color = _read_param(params, 'background_color', 'white')
        include_sha1sum = _read_param(params, 'include_sha1sum', False)
        print_progress = _read_param(params, 'print_progress', False)

        text = self.video.format_metadata(params={
            'include_sha1sum': include_sha1sum,
            'print_progress': print_progress,
        })

        _, total_height = draw_text_block(None, (0, 0), text, params={
            'font': text_font,
            'spacing': line_spacing,
            'dry_run': True,
        })

        metadata_sheet = Image.new('RGBA', (total_width, total_height),
                                   background_color)
        draw_text_block(metadata_sheet, (0, 0), text, params={
            'font': text_font,
            'color': text_color,
            'spacing': line_spacing,
        })

        return metadata_sheet

    @staticmethod
    def _gen_promotional_banner(total_width, params=None):
        """Generate promotion banner.

        This is the promotional banner for the storyboard package.

        Parameters
        ----------
        total_width : int
            Total width of the metadata sheet. Usually determined by the
            width of the bare storyboard.
        params : dict, optional
            Optional parameters enclosed in a dict. Default is
            ``None``. See the "Other Parameters" section for understood
            key/value pairs.

        Returns
        -------
        banner : PIL.Image.Image

        Other Parameters
        ----------------
        text_font : Font, optional
            Default is the font constructed by ``Font()`` without
            arguments.
        text_color: color, optional
            Default is 'black'.
        background_color: color, optional
            Default is 'white'.

        """

        if params is None:
            params = {}
        text_font = _read_param(params, 'text_font', Font())
        text_color = _read_param(params, 'text_color', 'black')
        background_color = _read_param(params, 'background_color', 'white')

        text = ("Generated by storyboard version %s. "
                "Fork me on GitHub: github.com/zmwangx/storyboard"
                % version.__version__)
        text_width, total_height = draw_text_block(None, (0, 0), text, params={
            'font': text_font,
            'dry_run': True,
        })

        banner = Image.new('RGBA', (total_width, total_height),
                           background_color)
        # center the text -- calculate the x coordinate of its topleft
        # corner
        text_x = int((total_width - text_width) / 2)
        draw_text_block(banner, (text_x, 0), text, params={
            'font': text_font,
            'color': text_color,
        })

        return banner


def main():
    """CLI interface."""

    # pylint: disable=too-many-statements,too-many-branches

    description = """Generate video storyboards with metadata reports.

    You may supply a list of videos. For each video, the generated
    storyboard image will be saved to a secure temporary file, and its
    absolute path will be printed to stdout for further manipulations
    (permanent archiving, uploading to an image hosting website,
    etc). Note that stdout is guaranteed to only receive the image
    paths, one per line, so you may embed this program in a streamlined
    script; stderr, on the other hand, may receive progress information
    without guaranteed format (see the --print-progress option).

    Below is the list of available options and their brief
    explanations. The options can also be stored in a configuration
    file, $XDG_CONFIG_HOME/storyboard/storyboard.conf (or if
    $XDG_CONFIG_HOME is not defined,
    ~/.config/storyboard/storyboard.conf), under the "storyboard-cli"
    section (the conf file format follows that described in
    https://docs.python.org/3/library/configparser.html).

    Note that the storyboard is in fact much more customizable; see the
    API reference of
    storyboard.storyboard.StoryBoard.gen_storyboard. Those customization
    parameters are not exposed in the CLI, but you may easily write a
    wrapper script around the storyboard.storyboard if you'd like to.

    For more detailed explanations, see
    http://storyboard.rtfd.org/en/stable/storyboard-cli.html (or replace
    "stable" with the version you are using).
    """
    parser = argparse.ArgumentParser(description=description)
    parser.add_argument(
        'videos', nargs='+', metavar='VIDEO',
        help="Path(s) to the video file(s).")
    parser.add_argument(
        '--ffmpeg-bin', metavar='NAME',
        help="""The name/path of the ffmpeg binary. The binay is
        guessed from OS type if this option is not specified.""")
    parser.add_argument(
        '--ffprobe-bin', metavar='NAME',
        help="""The name/path of the ffprobe binary. The binay is
        guessed from OS type if this option is not specified.""")
    parser.add_argument(
        '-f', '--output-format', choices=['jpeg', 'png'],
        help="Output format of the storyboard image. Default is JPEG.")
    parser.add_argument(
        '--quality', type=int,
        help="""Quality of the output image, should be an integer
        between 1 and 100. Only meaningful when the output format is
        JPEG. Default is 85.""")
    parser.add_argument(
        '--video-duration', type=float, metavar='SECONDS',
        help="""Video duration in seconds (float). By default the
        duration is extracted from container metadata, but in case it is
        not available or wrong, use this option to correct it and get a
        saner storyboard. Note however that this option activates output
        seeking (i.e., seeking the video frame by frame) in thumbnail
        generation, so it will be *infinitely* slower than without this
        option.""")
    parser.add_argument(
        '--exclude-sha1sum', '-s', action='store_const', const=True,
        help="Exclude SHA-1 digest of the video(s) from storyboard(s).")
    parser.add_argument(
        '--include-sha1sum', action='store_true',
        help="""Include SHA-1 digest of the video(s). Overrides
        '--exclude-sha1sum'. This option is only useful if
        exclude_sha1sum is turned on by default in the config file.""")
    parser.add_argument(
        '--verbose', '-v', choices=['auto', 'on', 'off'],
        nargs='?', const='auto',
        help="""Whether to print progress information to stderr. Default
        is 'auto'.""")
    parser.add_argument(
        '--version', action='version', version=version.__version__)
    cli_args = parser.parse_args()

    if 'XDG_CONFIG_HOME' in os.environ:
        config_file = os.path.join(os.environ['XDG_CONFIG_HOME'],
                                   'storyboard/storyboard.conf')
    else:
        config_file = os.path.expanduser('~/.config/storyboard/storyboard.conf')

    ffmpeg_bin_guessed, ffprobe_bin_guessed = fflocate.guess_bins()
    defaults = {
        'ffmpeg_bin': ffmpeg_bin_guessed,
        'ffprobe_bin': ffprobe_bin_guessed,
        'output_format': 'jpeg',
        'quality': 85,
        'video_duration': None,
        'exclude-sha1sum': False,
        'verbose': 'auto',
    }

    optreader = util.OptionReader(
        cli_args=cli_args,
        config_files=config_file,
        section='storyboard-cli',
        defaults=defaults,
    )
    bins = (optreader.opt('ffmpeg_bin'), optreader.opt('ffprobe_bin'))
    output_format = optreader.opt('output_format')
    if output_format not in ['jpeg', 'png']:
        msg = ("fatal error: output format should be either 'jpeg' or 'png'; "
               "'%s' received instead\n" % output_format)
        sys.stderr.write(msg)
        exit(1)
    suffix = '.jpg' if output_format == 'jpeg' else '.png'
    quality = optreader.opt('quality', opttype=int)
    video_duration = optreader.opt('video_duration', opttype=float)
    include_sha1sum = not optreader.opt('exclude_sha1sum', opttype=bool)
    if cli_args.include_sha1sum:
        # force override
        include_sha1sum = True
    verbose = optreader.opt('verbose')
    if verbose == 'on':
        print_progress = True
    elif verbose == 'off':
        print_progress = False
    else:
        if verbose != 'auto':
            msg = ("warning: '%s' is a not a valid argument to --verbose; "
                   "ignoring and using 'auto' instead\n" % verbose)
            sys.stderr.write(msg)
        if sys.stderr.isatty():
            print_progress = True
        else:
            print_progress = False

    # test bins
    try:
        fflocate.check_bins(bins)
    except OSError:
        msg = ("fatal error: at least one of '%s' and '%s' does not exist on "
               "PATH or is corrupted (expected ffmpeg and ffprobe)\n" % bins)
        sys.stderr.write(msg)
        exit(1)

    # real stuff happens from here
    returncode = 0
    for video in cli_args.videos:
        try:
            storyboard_image = StoryBoard(video, params={
                'bins': bins,
                'video_duration': video_duration,
                'print_progress': print_progress,
            }).gen_storyboard(params={
                'include_sha1sum': include_sha1sum,
                'print_progress': print_progress,
            })
        except OSError as err:
            sys.stderr.write("error: %s\n\n" % str(err))
            returncode = 1
            continue

        tempfd, storyboard_file = tempfile.mkstemp(
            prefix='storyboard-', suffix=suffix)
        os.close(tempfd)
        if output_format == 'jpeg':
            storyboard_image.save(storyboard_file, 'jpeg', quality=quality,
                                  optimize=True, progressive=True)
        else:  # 'png'
            storyboard_image.save(storyboard_file, 'png', optimize=True)

        if print_progress:
            sys.stderr.write("\n")
            sys.stderr.write("storyboard saved to: ")
            sys.stderr.flush()
            print(storyboard_file)
            sys.stderr.write("\n")
        else:
            print(storyboard_file)
    return returncode


if __name__ == "__main__":
    exit(main())



#!/usr/bin/env python3

"""Extract video frames.

Classes
-------
.. autosummary::
    Frame

Routines
--------
.. autosummary::
    extract_frame

----

"""

from __future__ import absolute_import
from __future__ import print_function

import io
import os
import subprocess

from PIL import Image

from storyboard import fflocate
from storyboard.util import read_param as _read_param


class Frame(object):
    """Video frame object containing a timestamp and an image.

    Parameters
    ----------
    timestamp : float
        Timestamp of the frame, in seconds.
    image : PIL.Image.Image
        Image of the frame.

    Attributes
    ----------
    timestamp : float
    image : PIL.Image.Image

    """

    # pylint: disable=too-few-public-methods

    def __init__(self, timestamp, image):
        assert isinstance(timestamp, int) or isinstance(timestamp, float),\
            "timestamp is not an int or float"
        assert isinstance(image, Image.Image),\
            "image is not a PIL.Image.Image instance"
        self.timestamp = timestamp
        self.image = image


def extract_frame(video_path, timestamp, params=None):
    """Extract a video frame from a given timestamp.

    Use FFmpeg to seek to the specified timestamp and decode the
    corresponding frame.

    Parameters
    ----------
    video_path : str
        Path to the video file.
    timestamp : float
        Timestamp in seconds (as a nonnegative float).
    params : dict, optional
        Optional parameters enclosed in a dict. Default is ``None``.
        See the "Other Parameters" section for understood key/value
        pairs.

    Returns
    -------
    frame : Frame

    Raises
    ------
    OSError
        If video file doesn't exist, ffmpeg binary doesn't exist or
        fails to run, or ffmpeg runs but generates no output (possibly
        due to an out of range timestamp).

    Other Parameters
    ----------------
    ffmpeg_bin : str, optional
        Name or path of FFmpeg binary. If ``None``, make educated guess
        using ``storyboard.fflocate.guess_bins``. Default is ``None``.
    codec : str, optional
        Image codec used by FFmpeg when outputing the frame. Default is
        ``'png'``. There is no need to touch this option unless your
        FFmpeg cannot encode PNG, which is very unlikely.
    frame_by_frame : bool, optional
        Whether to seek frame by frame, i.e., whether to use output
        seeking (see https://trac.ffmpeg.org/wiki/Seeking). Default is
        ``False``. Note that seeking frame by frame is *extremely* slow,
        but accurate. Only use this when the container metadata is wrong
        or missing, so that input seeking produces wrong image.

    """

    if params is None:
        params = {}
    if 'ffmpeg_bin' in params and params['ffmpeg_bin'] is not None:
        ffmpeg_bin = params['ffmpeg_bin']
    else:
        ffmpeg_bin, _ = fflocate.guess_bins()
    codec = _read_param(params, 'codec', 'png')
    frame_by_frame = (params['frame_by_frame'] if 'frame_by_frame' in params
                      else False)

    if not os.path.exists(video_path):
        raise OSError("video file '%s' does not exist" % video_path)

    ffmpeg_args = [ffmpeg_bin]
    if frame_by_frame:
        # output seeking
        ffmpeg_args += [
            '-i', video_path,
            '-ss', str(timestamp),
        ]
    else:
        # input seeking
        ffmpeg_args += [
            '-ss', str(timestamp),
            '-i', video_path,
        ]
    ffmpeg_args += [
        '-f', 'image2',
        '-vcodec', codec,
        '-vframes', '1',
        '-hide_banner',
        '-',
    ]
    proc = subprocess.Popen(ffmpeg_args,
                            stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    frame_bytes, ffmpeg_err = proc.communicate()
    if proc.returncode != 0:
        msg = (("ffmpeg failed to extract frame at time %.2f\n"
                "ffmpeg error message:\n%s") %
               (timestamp, ffmpeg_err.strip().decode('utf-8')))
        raise OSError(msg)

    if not frame_bytes:
        # empty output, no frame generated
        msg = ("ffmpeg generated no output "
               "(timestamp %.2f might be out of range)"
               "ffmpeg error message:\n%s" %
               (timestamp, ffmpeg_err.strip().decode('utf-8')))
        raise OSError(msg)

    try:
        frame_image = Image.open(io.BytesIO(frame_bytes))
    except IOError:
        raise OSError("failed to open frame with PIL.Image.open")

    return Frame(timestamp, frame_image)
